/*
Module : GPS.CPP
Purpose: "C" style Implementation to GPS32
Created: PJN / 28-12-1997
History: PJN / 20-01-1998  1.  A number of problems were discovered when GPSLIB was being
                               used in a UNICODE build. These problems have been fixed
                           2.  Mak file now includes all the possible build options
                           3.  Mak file no longer hard codes output to d:\winnt\system32
                           4.  Included release binaries in distribution
			   PJN / 26-08-1998  1.  Small change to stdafx.h to include afxcview.h in gps32exe
				                   2.  The GpsShowControlPanel function now actually works
													 3.  Minor tweaks to the GPS Mak files
													 4.  Removed a number of unused symbols
													 5.  Change ID of a number of GPS32EXE menu items to avoid conflict
													     with new VC 6 defines.
													 6.  Updated all Version Infos to 1.02
													 7.  Removed a level 4 warning from the Test app.
													 8.  Removed unused toolbar from GPS CPL file
													 9.  GPS cpl file now uses standard F2 accelerator for Rename
                           10. GPS32EXE toolbar is now shown flat
													 11. Number of extra TRACE statements have been added to 
													     aid in debugging.
         PJN / 11-08-1999  1.  Now uses the WINAPI calling convention meaning that GPSLIB is now
                               callable from VB.
                           2.  Now also ships a GPS.BAS file for use of GPSLIB in VB
         PJN / 27-09-2001  1.  Fixed a problem where the background serial port thread using causing
                           heavy CPU utilization.
                           2.  Updated the project settings so that all binaries get build to the Bin
                               directory and all Lib files get created in a Lib directory
                           3.  Updated the version info's in all the binaries
                           4.  Updated the copyright message in the code modules
         PJN / 21-08-2002  1.  Fixed a synchronisation problem between the worker thread and the 
                           GpsGetPosition method. Thanks to Bill Oatman for spotting this problem.



Copyright (c) 1997 - 2002 by PJ Naughter.  (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////////  Includes  //////////////////////////////////
#include "stdafx.h"
#include "gps.h"
#include "serport.h"
#include "resource.h"
#include "nmea.h"
#include "math.h"
#include "gpssetup.h"



//////////////////////////////////  Macros / Defines //////////////////////////
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif


////////////////////////////////// Locals /////////////////////////////////////
class GPSHandle
{
public:
  GPSHandle();
  ~GPSHandle();

  GPSPOSITION      m_Position;    //the actual data
  CEvent*          m_pKillEvent;  //Event to signal thread to exit
  CEvent*          m_pStartEvent; //Is the background thread running
  GPSDEVINFO       m_DevInfo;     //Settings of the comms port to use
  BOOL             m_bRunning;    //Is the background thread running
  CWinThread*      m_pThread;     //The pointer to the background thread
  CCriticalSection m_csPosition;  //Critical section used to serialized access to m_Position
};


BOOL GetGpsDevice(DWORD dwDevice, LPGPSDEVINFO lpGpsDevInfo);
BOOL SetGpsDevice(DWORD dwDevice, LPCGPSDEVINFO lpGpsDevInfo, BOOL bCheckDefault = TRUE);
BOOL SetGpsNumDevices(DWORD dwDevices);
UINT GpsMonitorThead(LPVOID pParam);


////////////////////////////////// Implementation /////////////////////////////

GPSHandle::GPSHandle()
{
  ZeroMemory(&m_Position, sizeof(GPSPOSITION));
  m_pKillEvent = new CEvent(FALSE, TRUE);
  m_pStartEvent = new CEvent(FALSE, TRUE);
  m_pThread = NULL;
}


GPSHandle::~GPSHandle()
{
  delete m_pStartEvent;
  m_pStartEvent = NULL;

  delete m_pKillEvent;
  m_pKillEvent = NULL;
} 


BOOL GetGpsDevice(DWORD dwDevice, LPGPSDEVINFO lpGpsDevInfo)
{
  BOOL bSuccess = FALSE;
  HKEY hKey;

  CString sValueKey;
  sValueKey.Format(_T("SOFTWARE\\PJ Naughter\\GPS32\\%d"), dwDevice);

  LONG nError = RegOpenKeyEx(HKEY_LOCAL_MACHINE, sValueKey, 0, KEY_ALL_ACCESS, &hKey);
  if (nError == ERROR_SUCCESS) 
  {
    DWORD dwType;
    DWORD dwSize;

    RegQueryValueEx(hKey, _T("DeviceName"), 0, &dwType, NULL, &dwSize);
    TCHAR* pszName = new TCHAR[dwSize / sizeof(TCHAR)];
    RegQueryValueEx(hKey, _T("DeviceName"), 0, &dwType, (LPBYTE) pszName, &dwSize);
    _tcscpy(lpGpsDevInfo->szDeviceName, pszName);
    delete [] pszName;

    RegQueryValueEx(hKey, _T("DefaultReceiver"), 0, &dwType, (LPBYTE) &lpGpsDevInfo->bDefaultReceiver, &dwSize);

    RegQueryValueEx(hKey, _T("Port"), 0, &dwType, (LPBYTE) &lpGpsDevInfo->wCommPort, &dwSize);

    RegQueryValueEx(hKey, _T("BaudRate"), 0, &dwType, (LPBYTE) &lpGpsDevInfo->dwCommBaudRate, &dwSize);

    RegQueryValueEx(hKey, _T("DataBits"), 0, &dwType, (LPBYTE) &lpGpsDevInfo->wCommDataBits, &dwSize);

    RegQueryValueEx(hKey, _T("StopBits"), 0, &dwType, (LPBYTE) &lpGpsDevInfo->wCommStopBits, &dwSize);

    bSuccess = TRUE;

    RegCloseKey(hKey);
  }
	else
	{
	  TRACE(_T("GetGpsDevice, Failed to open a registry key, Error was: %d\n"), nError);
	}

  return bSuccess;
}


BOOL GpsSetNumDevices(DWORD dwDevices)
{
  BOOL bSuccess = FALSE;
  HKEY hKey;
  DWORD dwDisposition;

  LONG nError = RegCreateKeyEx(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\PJ Naughter\\GPS32"), 0, _T(""),
                               REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, &dwDisposition);
  if (nError == ERROR_SUCCESS) 
  {
    RegSetValueEx(hKey, _T("NumberOfDevices"), 0, REG_DWORD, (CONST BYTE*) &dwDevices, sizeof(DWORD));
    
    RegCloseKey(hKey);

    bSuccess = TRUE;
  }
	else
	{
    TRACE(_T("GpsSetNumDevices, Failed in call to create a registry key, Error was: %d\n"), nError);
	}

  return bSuccess;
}


BOOL SetGpsDevice(DWORD dwDevice, LPCGPSDEVINFO lpDevice, BOOL bCheckDefault)
{
  //Fix up the case where we have just made this device 
  //the default by removing the default flag from all 
  //other devices
  if (bCheckDefault && lpDevice->bDefaultReceiver)
  {
    //make all other devices non default
    for (DWORD i=0; i<GpsGetNumDevices(); i++)
    {
      GPSDEVINFO devInfo;
      GetGpsDevice(i, &devInfo);
      devInfo.bDefaultReceiver = FALSE; 
      SetGpsDevice(i, &devInfo, FALSE);
    }
  }

  BOOL bSuccess = FALSE;
  HKEY hKey;
  DWORD dwDisposition;
  CString sValueKey;
  sValueKey.Format(_T("SOFTWARE\\PJ Naughter\\GPS32\\%d"), dwDevice);
	LONG nError = RegCreateKeyEx(HKEY_LOCAL_MACHINE, sValueKey, 0, _T(""), REG_OPTION_NON_VOLATILE,
                               KEY_ALL_ACCESS, NULL, &hKey, &dwDisposition);
  if (nError == ERROR_SUCCESS) 
  {
    RegSetValueEx(hKey, _T("DeviceName"), 0, REG_SZ, (CONST BYTE*) lpDevice->szDeviceName, _tcslen(lpDevice->szDeviceName));

    RegSetValueEx(hKey, _T("DefaultReceiver"), 0, REG_DWORD, (CONST BYTE*) &lpDevice->bDefaultReceiver, sizeof(BOOL));

    DWORD dwData = lpDevice->wCommPort;
    RegSetValueEx(hKey, _T("Port"), 0, REG_DWORD, (CONST BYTE*) &dwData, sizeof(DWORD));

    RegSetValueEx(hKey, _T("BaudRate"), 0, REG_DWORD, (CONST BYTE*) &lpDevice->dwCommBaudRate, sizeof(DWORD));

    dwData = lpDevice->wCommDataBits;
    RegSetValueEx(hKey, _T("DataBits"), 0, REG_DWORD, (CONST BYTE*) &dwData, sizeof(DWORD));

    dwData = lpDevice->wCommStopBits;
    RegSetValueEx(hKey, _T("StopBits"), 0, REG_DWORD, (CONST BYTE*) &dwData, sizeof(DWORD));

    RegCloseKey(hKey);

    bSuccess = TRUE;
  }
	else
	{
    TRACE(_T("SetGPSDevice, Failed in call to create a registry key, Error was: %d\n"), nError);
	}

  //make sure that at least one device is default
  if (bCheckDefault && !lpDevice->bDefaultReceiver)
  {
    DWORD dwDevices = GpsGetNumDevices();
    GPSDEVINFO* pGpsDevInfo = new GPSDEVINFO[dwDevices];
    dwDevices = GpsEnumDevices(pGpsDevInfo, dwDevices);
    BOOL bFound = FALSE;
    for (DWORD i=0; i<dwDevices && !bFound; i++)
    {
      if (pGpsDevInfo[i].bDefaultReceiver)
        bFound = TRUE;
    }
    delete [] pGpsDevInfo;

    if (!bFound)
    {
      TRACE(_T("SetGpsDevice, Found no device with default receiver attribute\n"));
      TRACE(_T("  making first device the default\n"));

      GPSDEVINFO devInfo;
      GetGpsDevice(0, &devInfo);
      devInfo.bDefaultReceiver = TRUE; 
      SetGpsDevice(0, &devInfo, FALSE);
    }
  }

  return bSuccess;
}


BOOL WINAPI GpsSetDevice(LPCTSTR lpszEntry, LPCGPSDEVINFO lpGpsDevInfo)
{
  //Find the device with the corresponding entry name
  DWORD dwDevices = GpsGetNumDevices();
  GPSDEVINFO* pGpsDevInfo = new GPSDEVINFO[dwDevices];
  dwDevices = GpsEnumDevices(pGpsDevInfo, dwDevices);
  BOOL bFound = FALSE;
  for (DWORD i=0; i<dwDevices && !bFound; i++)
  {
    if (_tcsicmp(lpszEntry, pGpsDevInfo[i].szDeviceName) == 0)
      bFound = TRUE;
  }
  delete [] pGpsDevInfo;

  if (!bFound)
  {
    TRACE(_T("GpsSetDevice, Failed to find GpsEntry for %s\n"), lpszEntry);
    return FALSE;
  }

  return SetGpsDevice(i-1, lpGpsDevInfo);
}


BOOL WINAPI GpsShowControlPanel()
{
#ifdef _DEBUG
  #ifdef _UNICODE
  	UINT nExec = ::WinExec("GPS103UD.EXE", SW_SHOW);
  #else
		UINT nExec = ::WinExec("GPS103D.EXE", SW_SHOW);
  #endif
#else
  #ifdef _UNICODE
  	UINT nExec = ::WinExec("GPS103U.EXE", SW_SHOW);
  #else
		UINT nExec = ::WinExec("GPS103.EXE", SW_SHOW);
	#endif
#endif
  if (nExec <= 31)
	  TRACE(_T("GpsShowControlPanel, Failed in call to WinExec GPS executable, WinExec returns: %d"), nExec);
  
	return (nExec > 31);
}


BOOL WINAPI GpsCreateEntry(HWND hWnd)
{
  BOOL bSuccess = FALSE;

  CWnd* pParent = CWnd::FromHandle(hWnd);

	CInstallPropertySheet propSheet(IDS_GPS_SETUP, pParent);
  propSheet.m_psh.dwFlags |= PSH_NOAPPLYNOW;
	int nResponse = propSheet.DoModal();

  if (nResponse == ID_WIZFINISH)
  {
    DWORD dwDevices = GpsGetNumDevices();

    GPSDEVINFO GpsInfo;
    _tcscpy(GpsInfo.szDeviceName, propSheet.m_Page3.m_sName); 
    GpsInfo.bDefaultReceiver = (propSheet.m_Page3.m_nMakeDefault == 0);
    GpsInfo.wCommPort = (WORD) propSheet.m_Page2.m_dwPort;
    GpsInfo.dwCommBaudRate = propSheet.m_Page2.m_dwBaudRate;
    GpsInfo.wCommDataBits = 8; 
    GpsInfo.wCommParity = GpsParityNone;
    GpsInfo.wCommStopBits = GpsStopBits1;

    bSuccess = SetGpsDevice(dwDevices, &GpsInfo);

    //increment the number of devices parameter
    GpsSetNumDevices(dwDevices + 1);
  }

  return bSuccess;
}


DWORD WINAPI GpsGetNumDevices()
{
  DWORD dwDevices = 0;
  DWORD dwType;
  DWORD dwSize = sizeof(DWORD);
  HKEY hKey;

  LONG nError = RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\PJ Naughter\\GPS32"),
                             0, KEY_ALL_ACCESS, &hKey);
  if (nError == ERROR_SUCCESS) 
  {
    RegQueryValueEx(hKey, _T("NumberOfDevices"), 0, &dwType, (LPBYTE) &dwDevices, &dwSize);
    RegCloseKey(hKey);
  }
	else
	{
    TRACE(_T("GpsGetNumDevices, Failed in call to open the registry, Error was: %d\n"), nError);
	}


  return dwDevices;
}


DWORD WINAPI GpsEnumDevices(LPGPSDEVINFO lpRasDevInfo, DWORD dwRequestedDevices)
{
  DWORD dwDevicesToRetreive = min(GpsGetNumDevices(), dwRequestedDevices);

  for (DWORD i=0; i<dwDevicesToRetreive; i++)
    GetGpsDevice(i, &lpRasDevInfo[i]);

  return dwDevicesToRetreive;
}


BOOL WINAPI GpsDeleteEntry(LPCTSTR lpszEntry)
{
  //Find the device with the corresponding entry name
  DWORD dwDevices = GpsGetNumDevices();
  GPSDEVINFO* pGpsDevInfo = new GPSDEVINFO[dwDevices];
  dwDevices = GpsEnumDevices(pGpsDevInfo, dwDevices);
  BOOL bFound = FALSE;
  for (DWORD i=0; i<dwDevices && !bFound; i++)
  {
    if (_tcsicmp(lpszEntry, pGpsDevInfo[i].szDeviceName) == 0)
      bFound = TRUE;
  }
  delete [] pGpsDevInfo;

  if (!bFound)
  {
    TRACE(_T("GpsDeleteEntry, Failed to find GpsEntry for %s\n"), lpszEntry);
    return FALSE;
  }

  //resequence all devices above the deleted item by
  //moving each item down by one
  GPSDEVINFO devInfo;
  for (DWORD j=i; j<dwDevices; j++)
  {
    GetGpsDevice(j, &devInfo);
    SetGpsDevice(j-1, &devInfo);
  }

  //Decrement the number of devices by one
  GpsSetNumDevices(dwDevices-1);

  return bFound;
}


BOOL WINAPI GpsRenameEntry(LPCTSTR lpszOldEntry, LPCTSTR lpszNewEntry)
{
  if (GpsDeviceNameAlreadyExists(lpszNewEntry))
  {
    TRACE(_T("GpsRename, Cannot rename as entry already exists with new name %s\n"), lpszNewEntry);
    return FALSE;
  }

  //Find the device with the corresponding entry name
  DWORD dwDevices = GpsGetNumDevices();
  GPSDEVINFO* pGpsDevInfo = new GPSDEVINFO[dwDevices];
  dwDevices = GpsEnumDevices(pGpsDevInfo, dwDevices);
  BOOL bFound = FALSE;
  for (DWORD i=0; i<dwDevices && !bFound; i++)
  {
    if (_tcsicmp(lpszOldEntry, pGpsDevInfo[i].szDeviceName) == 0)
      bFound = TRUE;
  }
  delete [] pGpsDevInfo;

  if (!bFound)
  {
    TRACE(_T("GpsRenameEntry, Failed to find GpsEntry for %s\n"), lpszOldEntry);
    return FALSE;
  }

  GPSDEVINFO devInfo;
  GetGpsDevice(i-1, &devInfo);
  _tcscpy(devInfo.szDeviceName, lpszNewEntry); 
  SetGpsDevice(i-1, &devInfo);

  return bFound;
}


BOOL WINAPI GpsPropertiesDlg(HWND hWnd, LPCTSTR lpszEntry)
{
  //Find the device with the corresponding entry name
  GPSDEVINFO devInfo;
  if (!GpsGetDevice(lpszEntry, &devInfo))
  {
    TRACE(_T("GpsPropertiesDlg, Failed to find device\n"));
    return FALSE;
  }

  CWnd* pParent = CWnd::FromHandle(hWnd);

  CString sCaption;
  AfxFormatString1(sCaption, IDS_PROPERTIES_CAPTION, lpszEntry);

	CPropertiesPropertySheet propSheet(sCaption, pParent);
  propSheet.m_psh.dwFlags |= PSH_NOAPPLYNOW;

  propSheet.m_Page1.m_bDefault = devInfo.bDefaultReceiver;
  propSheet.m_Page1.m_dwStopBits = devInfo.wCommStopBits;
  propSheet.m_Page1.m_dwPort = devInfo.wCommPort;
  propSheet.m_Page1.m_dwParity = devInfo.wCommParity;
  propSheet.m_Page1.m_dwBaudRate = devInfo.dwCommBaudRate;
  propSheet.m_Page1.m_dwDataBits = devInfo.wCommDataBits;

	int nResponse = propSheet.DoModal();

  if (nResponse == IDOK)
  {
    devInfo.bDefaultReceiver = propSheet.m_Page1.m_bDefault;
    devInfo.wCommStopBits    = (WORD) propSheet.m_Page1.m_dwStopBits;
    devInfo.wCommPort        = (WORD) propSheet.m_Page1.m_dwPort;
    devInfo.wCommParity      = (WORD) propSheet.m_Page1.m_dwParity;
    devInfo.dwCommBaudRate   = propSheet.m_Page1.m_dwBaudRate;
    devInfo.wCommDataBits    = (WORD) propSheet.m_Page1.m_dwDataBits;

    if (!GpsSetDevice(lpszEntry, &devInfo))
    {
      TRACE(_T("GpsPropertiesDlg, Failed to set device paramters\n"));
      return FALSE;
    }

    return TRUE;
  }
  else
    return FALSE;
}



HGPS WINAPI GpsOpen(LPCTSTR lpszEntry)
{
  //check to make sure the device to open exists
  GPSDEVINFO devInfo;
  if (!GpsGetDevice(lpszEntry, &devInfo))
  {
    TRACE(_T("GpsOpen, Failed to find device\n"));
    return (HGPS) INVALID_HANDLE_VALUE;
  }

  //create a new handle instance
  GPSHandle* pGpsHandle = new GPSHandle;

  //In the future we will support sharing of GPS devices across process boundaries.
  //For the moment it will be 1 application at a time only.


  //this is the first client app to link to us, so spin of the
  //background thread which does the monitoring of the serial
  //port
  CopyMemory(&pGpsHandle->m_DevInfo, &devInfo, sizeof(devInfo));
  pGpsHandle->m_bRunning = TRUE;
  pGpsHandle->m_pThread = AfxBeginThread(GpsMonitorThead, pGpsHandle, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);
  if (!pGpsHandle->m_pThread)
  {
    TRACE(_T("GpsOpen, Failed to create background thread\n"));
    delete pGpsHandle;
    pGpsHandle = NULL;
    return (HGPS) INVALID_HANDLE_VALUE;
  }

  pGpsHandle->m_pThread->m_bAutoDelete = FALSE; //we're in charge of cleanup
  pGpsHandle->m_pThread->ResumeThread();

  //wait until the thread is correctly initialised
  WaitForSingleObject(*pGpsHandle->m_pStartEvent, INFINITE);

  if (!pGpsHandle->m_bRunning)
  {
    //wait for the thread to exit
    WaitForSingleObject(pGpsHandle->m_pThread->m_hThread, INFINITE);

    delete pGpsHandle->m_pThread;
    pGpsHandle->m_pThread = NULL;

    delete pGpsHandle;
    pGpsHandle = NULL;

    TRACE(_T("GpsOpen, Failed to start background thread\n"));

    return (HGPS) INVALID_HANDLE_VALUE;
  }

  return (HGPS) pGpsHandle;
}


BOOL WINAPI GpsClose(HGPS hEntry)
{
  if (hEntry == INVALID_HANDLE_VALUE)
  {
    TRACE(_T("GpsClose, invalid HGPS handle value\n"));
    return FALSE;
  }

  GPSHandle* pGpsHandle = (GPSHandle*) hEntry;
  if (pGpsHandle == NULL)
  {
    TRACE(_T("GpsClose, invalid HGPS handle value\n"));
    return FALSE;
  }

  pGpsHandle->m_pKillEvent->SetEvent();
  WaitForSingleObject(pGpsHandle->m_pThread->m_hThread, INFINITE);

  delete pGpsHandle->m_pThread;
  pGpsHandle->m_pThread = NULL;

  delete pGpsHandle;
  pGpsHandle = NULL;

  return TRUE;
}


BOOL WINAPI GpsGetPosition(HGPS hEntry, LPGPSPOSITION lpPosition)
{
  if (hEntry == INVALID_HANDLE_VALUE)
  {
    TRACE(_T("GpsGetPostion, invalid HGPS handle value\n"));
    return FALSE;
  }

  GPSHandle* pGpsHandle = (GPSHandle*) hEntry;
  if (pGpsHandle == NULL)
  {
    TRACE(_T("GpsGetPostion, invalid HGPS handle value\n"));
    return FALSE;
  }

  CSingleLock sl(&pGpsHandle->m_csPosition, TRUE); //serialize access to the position data
  CopyMemory(lpPosition, &pGpsHandle->m_Position, sizeof(GPSPOSITION));

  return TRUE;
}


DWORD WINAPI GpsGetVersion()
{
  return 0x0103;  //ie 1.03

  /* Revision History

  v1.0   29/12/1997 Original Release
  v1.01  20/01/1998 Number of problems fixed as described at top of this file
	v1.02  26/08/1998 Number of problems/updates as described at top of this file
  v1.03  11/06/1999 Now uses the WINAPI (_stdcall) calling convention

  */
}

BOOL WINAPI GpsGetDevice(LPCTSTR lpszEntry, LPGPSDEVINFO lpGpsDevInfo)
{
  BOOL bFound = FALSE;
  if (lpszEntry == NULL) //look for the default device
  {
    DWORD dwDevices = GpsGetNumDevices();
    GPSDEVINFO* pGpsDevInfo = new GPSDEVINFO[dwDevices];
    dwDevices = GpsEnumDevices(pGpsDevInfo, dwDevices);
    for (DWORD i=0; i<dwDevices && !bFound; i++)
    {
      if (pGpsDevInfo[i].bDefaultReceiver)
      {
        CopyMemory(lpGpsDevInfo, &pGpsDevInfo[i], sizeof(pGpsDevInfo[i]));
        bFound = TRUE;
      }
    }
    delete [] pGpsDevInfo;

    if (!bFound)
      TRACE(_T("GpsGetDevice, Failed to find default device\n"));
  }
  else
  {
    DWORD dwDevices = GpsGetNumDevices();
    GPSDEVINFO* pGpsDevInfo = new GPSDEVINFO[dwDevices];
    dwDevices = GpsEnumDevices(pGpsDevInfo, dwDevices);
    for (DWORD i=0; i<dwDevices && !bFound; i++)
    {
      if (_tcsicmp(lpszEntry, pGpsDevInfo[i].szDeviceName) == 0)
      {
        bFound = TRUE;
        CopyMemory(lpGpsDevInfo, &pGpsDevInfo[i], sizeof(GPSDEVINFO));
      }
    }

    delete [] pGpsDevInfo;

    if (!bFound)
      TRACE(_T("GpsGetDevice, Failed to find device %s\n"), lpszEntry);
  }

  return bFound;
}


BOOL WINAPI GpsDeviceNameAlreadyExists(LPCTSTR lpszEntry)
{
  GPSDEVINFO devinfo;
  return GpsGetDevice(lpszEntry, &devinfo);
}


void WINAPI GpsShowAboutBox(HWND hWnd)
{
  CWnd* pParent = CWnd::FromHandle(hWnd);

  CDialog dlg(IDD_ABOUT, pParent);
  dlg.DoModal();
}


UINT GpsMonitorThead(LPVOID pParam)
{
  CString sSentence;
  CString sSerialString;
  GPSHandle* pGpsHandle = (GPSHandle*) pParam;
  BYTE szBuffer[100]; 

  CGpsSerialPort SerialPort;
  BOOL bOpen = SerialPort.Open(&pGpsHandle->m_DevInfo);
  if (!bOpen) 
  {
    TRACE(_T("GpsMonitorThread, Failed to open comms port\n"));
    pGpsHandle->m_bRunning = FALSE;
  }

  //thread is now initialised correctly, GpsOpen is free to continue
  pGpsHandle->m_pStartEvent->SetEvent();

  while (pGpsHandle->m_bRunning)
  {
    //if terminate event is active, terminate monitor thread 
    if (::WaitForSingleObject(pGpsHandle->m_pKillEvent->m_hObject, 0) == WAIT_OBJECT_0) 
    {
      pGpsHandle->m_bRunning = FALSE;
      continue;
    }
    
    int nBytesRead = SerialPort.Read(szBuffer, sizeof(szBuffer));

    if (nBytesRead > 0) 
    {
      int nIndex = 0;
      while (nIndex < nBytesRead)
      {
        sSerialString += ((TCHAR) szBuffer[nIndex]);
        nIndex++;
      }

      // See if we've got a sentence yet
      int nLocationLineFeed = sSerialString.Find(0x0A);

      while (nLocationLineFeed != -1) 
      {
        // Yes, there is a line feed in the string
        sSentence = sSerialString.Left(nLocationLineFeed + 1);
        // Now trim off the sentence from what came in from the serial port
        sSerialString = sSerialString.Right(sSerialString.GetLength() - (nLocationLineFeed+1));
        nLocationLineFeed = sSerialString.Find(0x0A);

        //determine which parser to call
        if (sSentence.Left(6).Compare(_T("$GPRMC")) == 0) 
        {
          CNmeaSentence sentence(sSentence);
          CRMCResponse response;
          if (response.Parse(sSentence) && response.m_IsDataValid)
          {
            CSingleLock sl(&pGpsHandle->m_csPosition, TRUE); //serialize access to the position data
            pGpsHandle->m_Position.dwLongitude = CNmeaSentence::NmeaAngleToHundrethsOfSeconds(response.m_Longitude.Value);
            pGpsHandle->m_Position.bEasting = (response.m_Longitude.Easting == East);
            pGpsHandle->m_Position.dwLatitude = CNmeaSentence::NmeaAngleToHundrethsOfSeconds(response.m_Latitude.Value);
            pGpsHandle->m_Position.bNorthing = (response.m_Latitude.Northing == North);
            pGpsHandle->m_Position.dwBearing = CNmeaSentence::NmeaBearingToHundrethsOfDegrees(response.m_Bearing);
            pGpsHandle->m_Position.dwSpeed = (DWORD) (response.m_SpeedOverGroundKnots * 185200);  //There's 1.8519998 Km/s in 1 Knot
            pGpsHandle->m_Position.wFixYear = response.m_Date.wYear;
            pGpsHandle->m_Position.wFixMonth = response.m_Date.wMonth;
            pGpsHandle->m_Position.wFixDay  = response.m_Date.wDay;
            pGpsHandle->m_Position.wFixHour = response.m_Time.wHour;
            pGpsHandle->m_Position.wFixMinute = response.m_Time.wMinute;
            pGpsHandle->m_Position.wFixSecond = response.m_Time.wSecond;
          }
        }
        else if (sSentence.Left(6).Compare(_T("$GPGGA")) == 0) 
        {
          CNmeaSentence sentence(sSentence);
          CGGAResponse response;
          if (response.Parse(sSentence))
          {
            CSingleLock sl(&pGpsHandle->m_csPosition, TRUE); //serialize access to the position data
            pGpsHandle->m_Position.nSatellites = (WORD) response.m_nSatellites;
            pGpsHandle->m_Position.dwAntennaAltitude = (DWORD) (response.m_AntennaAltitudeMeters * 100);
            pGpsHandle->m_Position.wQualityIndicator = (WORD) response.m_GPSQuality;
          }
        }
      }
    }

    //Sleep for a time before we loop around, this avoids heavy CPU utilization
    Sleep(100);
  }

  return 1;
}
